import unittest

import transaction
from BTrees.OOBTree import OOBTree
from zope.interface.verify import verifyObject
from zope.component import provideAdapter
from zope.app.folder import Folder
from zope.app.container.btree import BTreeContainer
from zope.app.testing import setup

from zodbbrowser.interfaces import IStateInterpreter, IObjectHistory
from zodbbrowser.history import ZodbObjectHistory
from zodbbrowser.btreesupport import (
        OOBTreeHistory,
        OOBTreeState,
        EmptyOOBTreeState, 
        FolderState,
        BTreeContainerState,
)
from zodbbrowser.tests.realdb import RealDatabaseTest


class TestOOBTreeState(unittest.TestCase):

    def setUp(self):
        setup.placelessSetUp()
        provideAdapter(ZodbObjectHistory)
        tree = OOBTree()
        tree[1] = 42
        tree[2] = 23
        tree[3] = 17
        state = tree.__getstate__()
        self.state = EmptyOOBTreeState(None, state, None)

    def tearDown(self):
        setup.placelessTearDown()

    def test_interface_compliance(self):
        verifyObject(IStateInterpreter, self.state)

    def test_getName(self):
        self.assertEquals(self.state.getName(), None)

    def test_getParent(self):
        self.assertEquals(self.state.getParent(), None)

    def test_listAttributes(self):
        self.assertEquals(self.state.listAttributes(), None)

    def test_listItems(self):
        self.assertEquals(list(self.state.listItems()),
                          [(1, 42), (2, 23), (3, 17)])

    def test_asDict(self):
        self.assertEquals(dict(self.state.asDict()), {1: 42, 2: 23, 3: 17})


class TestLargeOOBTreeState(RealDatabaseTest):

    def setUp(self):
        setup.placelessSetUp()
        provideAdapter(ZodbObjectHistory)
        RealDatabaseTest.setUp(self)
        self.tree = self.conn.root()['tree'] = OOBTree()
        for i in range(0, 1000):
            self.tree[i] = -1
        transaction.commit()
        for i in range(0, 1000, 2):
            self.tree[i] = 1
        transaction.commit()
        self.tids = [d['tid'] for d in IObjectHistory(self.tree)]

    def tearDown(self):
        RealDatabaseTest.tearDown(self)
        setup.placelessTearDown()

    def getState(self, tid):
        state = IObjectHistory(self.tree).loadState(tid)
        return OOBTreeState(self.tree, state, tid)

    def test_current_state(self):
        state = self.getState(None)
        self.assertEquals(sum(state.asDict().values()), 0)

    def test_historical_state(self):
        state = self.getState(self.tids[-1])
        self.assertEquals(sum(state.asDict().values()), -1000)

    def test_historical_state_does_not_leave_modified_caches(self):
        state = self.getState(self.tids[-1])
        self.assertEquals(sum(state.asDict().values()), -1000)
        self.assertEquals(sum(self.tree.values()), 0)


class TestLargeOOBTreeHistory(RealDatabaseTest):

    def setUp(self):
        setup.placelessSetUp()
        provideAdapter(ZodbObjectHistory)
        provideAdapter(OOBTreeHistory)
        RealDatabaseTest.setUp(self)
        self.tree = self.conn.root()['tree'] = OOBTree()
        for i in range(0, 100):
            self.tree[i] = i
            transaction.commit()

    def tearDown(self):
        RealDatabaseTest.tearDown(self)
        setup.placelessTearDown()

    def getState(self, tid):
        state = IObjectHistory(self.tree).loadState(tid)
        return OOBTreeState(self.tree, state, tid)

    def test_full_history(self):
        tids = [d['tid'] for d in IObjectHistory(self.tree)][::-1]
        for i in range(0, 100):
            state = self.getState(tids[i])
            self.assertEquals(len(state.asDict()), i + 1)

    def test_loadStatePickle(self):
        tids = [d['tid'] for d in IObjectHistory(self.tree)][::-1]
        for i in range(0, 100):
            # Regression test: make sure we're not getting those
            # POSKeyErrors.  LP#953480
            IObjectHistory(self.tree).loadStatePickle(tids[i])

    def test_rollback_to_last_state_does_nothing(self):
        history = IObjectHistory(self.tree)
        history.rollback(history.lastChange())
        self.assertEquals(len(self.tree), 100)
        # BTrees play funky games with cached lenghts, make sure the content
        # matches that
        self.assertEquals(len(list(self.tree)), 100)
        self.assertFalse(self.tree._p_changed)

    def test_rollback_changes_buckets(self):
        history = OOBTreeHistory(self.tree)
        history.rollback(history._lastRealChange())
        # therefore the state of the tree itself stays constant, but
        # one or more of its buckets change
        self.assertNotEquals(len(self.tree), 100)
        self.assertNotEquals(len(list(self.tree)), 100)
        self.assertFalse(self.tree._p_changed)

    def test_rollback(self):
        history = OOBTreeHistory(self.tree)
        tid = history[len(history) / 2]['tid']
        history.rollback(tid)
        self.assertEquals(len(self.tree), 50)
        # BTrees play funky games with cached lenghts, make sure the content
        # matches that
        self.assertEquals(len(list(self.tree)), 50)
        self.assertTrue(self.tree._p_changed)


class TestEmptyOOBTreeState(unittest.TestCase):

    def setUp(self):
        self.state = EmptyOOBTreeState(None, None, None)

    def test_interface_compliance(self):
        verifyObject(IStateInterpreter, self.state)

    def test_getName(self):
        self.assertEquals(self.state.getName(), None)

    def test_getParent(self):
        self.assertEquals(self.state.getParent(), None)

    def test_listAttributes(self):
        self.assertEquals(self.state.listAttributes(), None)

    def test_listItems(self):
        self.assertEquals(list(self.state.listItems()), [])

    def test_asDict(self):
        self.assertEquals(dict(self.state.asDict()), {})


class TestFolderState(RealDatabaseTest):

    def setUp(self):
        setup.placelessSetUp()
        provideAdapter(OOBTreeState)
        provideAdapter(ZodbObjectHistory)
        RealDatabaseTest.setUp(self)
        self.folder = self.conn.root()['folder'] = Folder()
        self.folder['foo'] = 1
        self.folder['bar'] = 2
        transaction.commit()
        self.state = FolderState(None, self.folder.__getstate__(),
                                 None)

    def tearDown(self):
        RealDatabaseTest.tearDown(self)
        setup.placelessTearDown()

    def test_listItems(self):
        self.assertEquals(list(self.state.listItems()),
                          [('bar', 2), ('foo', 1)])

    def test_listItems_no_data(self):
        state = FolderState(None, Folder().__getstate__(), None)
        self.assertEquals(list(state.listItems()), []);


class TestBTreeContainerState(RealDatabaseTest):

    def setUp(self):
        setup.placelessSetUp()
        provideAdapter(OOBTreeState)
        provideAdapter(ZodbObjectHistory)
        RealDatabaseTest.setUp(self)
        self.container = self.conn.root()['container'] = BTreeContainer()
        self.container['foo'] = 1
        self.container['bar'] = 2
        transaction.commit()
        self.state = BTreeContainerState(None, self.container.__getstate__(),
                                         None)

    def tearDown(self):
        RealDatabaseTest.tearDown(self)
        setup.placelessTearDown()

    def test_listItems(self):
        self.assertEquals(list(self.state.listItems()),
                          [('bar', 2), ('foo', 1)])

    def test_listItems_no_data(self):
        state = BTreeContainerState(None, BTreeContainer().__getstate__(),
                                    None)
        self.assertEquals(list(state.listItems()), []);




import logging

from persistent.dict import PersistentDict
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping
from zope.component import adapts, getMultiAdapter
from zope.interface import implements, Interface
from zope.interface.interfaces import IInterface
from zope.interface.interface import InterfaceClass
from zope.proxy import removeAllProxies
from zope.traversing.interfaces import IContainmentRoot
from ZODB.utils import u64

import zope.interface.declarations

# be compatible with Zope 3.4, but prefer the modern package structure
try:
    from zope.container.sample import SampleContainer
except ImportError:
    from zope.app.container.sample import SampleContainer # BBB
try:
    from zope.container.ordered import OrderedContainer
except ImportError:
    from zope.app.container.ordered import OrderedContainer # BBB
try:
    from zope.container.contained import ContainedProxy
except ImportError:
    from zope.app.container.contained import ContainedProxy # BBB

from zodbbrowser.interfaces import IStateInterpreter, IObjectHistory


log = logging.getLogger(__name__)


real_Provides = zope.interface.declarations.Provides


def install_provides_hack():
    """Monkey-patch zope.interface.Provides with a more lenient version.

    A common result of missing modules in sys.path is that you cannot
    unpickle objects that have been marked with directlyProvides() to
    implement interfaces that aren't currently available.  Those interfaces
    are replaced by persistent broken placeholders, which aren classes,
    not interfaces, and aren't iterable, causing TypeErrors during unpickling.
    """
    zope.interface.declarations.Provides = Provides


def flatten_interfaces(args):
    result = []
    for a in args:
        if isinstance(a, (list, tuple)):
            result.extend(flatten_interfaces(a))
        elif IInterface.providedBy(a):
            result.append(a)
        else:
            log.warning('  replacing %s with a placeholder', repr(a))
            result.append(InterfaceClass(a.__name__,
                            __module__='broken ' + a.__module__))
    return result


def Provides(cls, *interfaces):
    try:
        return real_Provides(cls, *interfaces)
    except TypeError, e:
        log.warning('Suppressing TypeError while unpickling Provides: %s', e)
        args = flatten_interfaces(interfaces)
        return real_Provides(cls, *args)


class ZodbObjectState(object):
    implements(IStateInterpreter)

    def __init__(self, obj, tid=None, _history=None):
        self.obj = removeAllProxies(obj)
        if _history is None:
            _history = IObjectHistory(self.obj)
        else:
            assert _history._obj is self.obj
        self.history = _history
        self.tid = None
        self.requestedTid = tid
        self.loadError = None
        self.pickledState = ''
        self._load()

    def _load(self):
        self.tid = self.history.lastChange(self.requestedTid)
        try:
            self.pickledState = self.history.loadStatePickle(self.tid)
            loadedState = self.history.loadState(self.tid)
        except Exception, e:
            self.loadError = "%s: %s" % (e.__class__.__name__, e)
            self.state = LoadErrorState(self.loadError, self.requestedTid)
        else:
            self.state = getMultiAdapter((self.obj, loadedState,
                                         self.requestedTid),
                                         IStateInterpreter)

    def getError(self):
        return self.loadError

    def listAttributes(self):
        return self.state.listAttributes()

    def listItems(self):
        return self.state.listItems()

    def getParent(self):
        return self.state.getParent()

    def getName(self):
        name = self.state.getName()
        if name is None:
            # __name__ is not in the pickled state, but it may be defined
            # via other means (e.g. class attributes, custom __getattr__ etc.)
            try:
                name = getattr(self.obj, '__name__', None)
            except Exception:
                # Ouch.  Oh well, we can't determine the name.
                pass
        return name

    def asDict(self):
        return self.state.asDict()

    # These are not part of IStateInterpreter

    def getObjectId(self):
        return u64(self.obj._p_oid)

    def isRoot(self):
        return IContainmentRoot.providedBy(self.obj)

    def getParentState(self):
        parent = self.getParent()
        if parent is None:
            return None
        else:
            return ZodbObjectState(parent, self.requestedTid)


class LoadErrorState(object):
    """Placeholder for when an object's state could not be loaded"""
    implements(IStateInterpreter)

    def __init__(self, error, tid):
        self.error = error
        self.tid = tid

    def getError(self):
        return self.error

    def getName(self):
        return None

    def getParent(self):
        return None

    def listAttributes(self):
        return []

    def listItems(self):
        return None

    def asDict(self):
        return {}


class GenericState(object):
    """Most persistent objects represent their state as a dict."""
    adapts(Interface, dict, None)
    implements(IStateInterpreter)

    def __init__(self, type, state, tid):
        self.state = state
        self.tid = tid

    def getError(self):
        return None

    def getName(self):
        return self.state.get('__name__')

    def getParent(self):
        return self.state.get('__parent__')

    def listAttributes(self):
        return self.state.items()

    def listItems(self):
        return None

    def asDict(self):
        return self.state


class PersistentMappingState(GenericState):
    """Convenient access to a persistent mapping's items."""
    adapts(PersistentMapping, dict, None)

    def listItems(self):
        return sorted(self.state.get('data', {}).items())


if PersistentMapping is PersistentDict:
    # ZODB 3.9 deprecated PersistentDict and made it an alias for
    # PersistentMapping.  I don't know a clean way to conditionally disable the
    # <adapter> directive in ZCML to avoid conflicting configuration actions,
    # therefore I'll register a decoy adapter registered for a decoy class.
    # This adapter will never get used.

    class DecoyPersistentDict(PersistentMapping):
        """Decoy to avoid ZCML errors while supporting both ZODB 3.8 and 3.9."""

    class PersistentDictState(PersistentMappingState):
        """Decoy to avoid ZCML errors while supporting both ZODB 3.8 and 3.9."""
        adapts(DecoyPersistentDict, dict, None)

else:
    class PersistentDictState(PersistentMappingState):
        """Convenient access to a persistent dict's items."""
        adapts(PersistentDict, dict, None)


class SampleContainerState(GenericState):
    """Convenient access to a SampleContainer's items"""
    adapts(SampleContainer, dict, None)

    def listItems(self):
        data = self.state.get('_SampleContainer__data')
        if not data:
            return []
        # data will be something persistent, maybe a PersistentDict, maybe a
        # OOBTree -- SampleContainer itself uses a plain Python dict, but
        # subclasses are supposed to overwrite the _newContainerData() method
        # and use something persistent.
        loadedstate = IObjectHistory(data).loadState(self.tid)
        return getMultiAdapter((data, loadedstate, self.tid),
                               IStateInterpreter).listItems()


class OrderedContainerState(GenericState):
    """Convenient access to an OrderedContainer's items"""
    adapts(OrderedContainer, dict, None)

    def listItems(self):
        # Now this is tricky: we want to construct a small object graph using
        # old state pickles without ever calling __setstate__ on a real
        # Persistent object, as _that_ would poison ZODB in-memory caches
        # in a nasty way (LP #487243).
        container = OrderedContainer()
        container.__setstate__(self.state)
        old_data_state = IObjectHistory(container._data).loadState(self.tid)
        old_order_state = IObjectHistory(container._order).loadState(self.tid)
        container._data = PersistentDict()
        container._data.__setstate__(old_data_state)
        container._order = PersistentList()
        container._order.__setstate__(old_order_state)
        return container.items()


class ContainedProxyState(GenericState):
    adapts(ContainedProxy, tuple, None)

    def __init__(self, proxy, state, tid):
        GenericState.__init__(self, proxy, state, tid)
        self.proxy = proxy

    def getName(self):
        return self.state[1]

    def getParent(self):
        return self.state[0]

    def listAttributes(self):
        return [('__name__', self.getName()),
                ('__parent__', self.getParent()),
                ('proxied_object', self.proxy.__getnewargs__()[0])]

    def listItems(self):
        return []

    def asDict(self):
        return dict(self.listAttributes())


class FallbackState(object):
    """Fallback when we've got no idea how to interpret the state"""
    adapts(Interface, Interface, None)
    implements(IStateInterpreter)

    def __init__(self, type, state, tid):
        self.state = state

    def getError(self):
        return None

    def getName(self):
        return None

    def getParent(self):
        return None

    def listAttributes(self):
        return [('pickled state', self.state)]

    def listItems(self):
        return None

    def asDict(self):
        return dict(self.listAttributes())




import os
import re
import sys
import tempfile
import shutil
import doctest
import unittest
import threading
from cgi import escape

import transaction
from lxml.html import fromstring, tostring
from ZODB.POSException import ReadOnlyError
from ZODB.FileStorage.FileStorage import FileStorage
from ZODB.DB import DB
from zope.testing.renormalizing import RENormalizing
from zope.testbrowser.browser import Browser
from zope.testbrowser.interfaces import IBrowser
from zope.app.testing import setup
from zope.app.publication.zopepublication import ZopePublication
from zope.app.folder.folder import Folder
from zope.app.appsetup.interfaces import DatabaseOpened
from zope.app.appsetup.bootstrap import bootStrapSubscriber

from zodbbrowser.standalone import main, serve_forever, stop_serving


class ServerController(object):

    def __init__(self):
        self.server_thread = None
        self.url = None
        self.port_number = None

    def findUnusedPort(self):
        """Find an unused TCP port number."""
        return 1234 # TODO: be smarter

    def run(self, *args):
        """Run the standalone ZODB Browser web app in the background."""
        if self.server_thread is not None:
            raise AssertionError('Already running a server')

        if not self.port_number:
            self.port_number = self.findUnusedPort()

        self.url = 'http://localhost:%d/' % self.port_number

        args += ('--listen', str(self.port_number))
        args += ('--quiet', )
        main(list(args), start_serving=False)

        self.server_thread = threading.Thread(name='server',
                                              target=serve_forever,
                                              kwargs=dict(interval=0.5))
        # Daemon threads are evil and cause weird errors on shutdown,
        # but we want ^C to not hang
        self.server_thread.setDaemon(True)
        self.server_thread.start()

    def stop(self):
        """Stop the background ZODB Browser web, if running."""
        if self.server_thread is not None:
            stop_serving()
            self.server_thread.join(1.0)
            self.server_thread = None
            self.url = None
            self.port_number = None
            # XXX: leaves zserver threads behind
            # XXX: leaves global Zope state, you'll need to run
            #      setup.placelessTearDown()


class TestsWithoutServer(object):
    """Functional tests for no special setup.

    Used simply for grouping purposes.
    """


class TestsWithServer(object):
    """Functional tests with a web app running in the background."""

    @classmethod
    def setUp(cls):
        cls.server = ServerController()
        cls.tempdir = tempfile.mkdtemp('zodbbrowser')
        cls.data_fs = os.path.join(cls.tempdir, 'data.fs')
        cls.createTestData()
        cls.server.run(cls.data_fs, '--rw')
        cls.url = cls.server.url

    @classmethod
    def tearDown(cls):
        cls.server.stop()
        shutil.rmtree(cls.tempdir)
        setup.placelessTearDown()

    @classmethod
    def createTestData(cls):
        storage = FileStorage(cls.data_fs)
        db = DB(storage)
        # Create root folder and all that jazz
        bootStrapSubscriber(DatabaseOpened(db))
        connection = db.open()
        root = connection.root()
        root_folder = root[ZopePublication.root_name]
        # This is not a great way to set up test fixtures, but it'll do
        # for now
        cls.createTestDataForBrowsing(root_folder)
        cls.createTestDataForRollbacking(root_folder)
        cls.createTestDataForRollbackCanBeCancelled(root_folder)
        connection.close()
        db.close()

    @classmethod
    def createTestDataForBrowsing(cls, root_folder):
        # set up data that browsing.txt expects
        root_folder['browsing'] = Folder()
        transaction.commit()

    @classmethod
    def createTestDataForRollbacking(cls, root_folder):
        # set up data that rollbacking.txt expects
        root_folder['rollbacking'] = Folder()
        transaction.commit()
        root_folder['rollbacking'].random_attribute = 'hey'
        transaction.commit()

    @classmethod
    def createTestDataForRollbackCanBeCancelled(cls, root_folder):
        # set up data that rollback-can-be-cancelled.txt expects
        root_folder['rbcbc'] = Folder()
        transaction.commit()
        root_folder['rbcbc'].random_attribute = 'hey'
        transaction.commit()


class TestCanCreateEmptyDataFs(unittest.TestCase):

    layer = TestsWithoutServer

    def setUp(self):
        self.tempdir = tempfile.mkdtemp('zodbbrowser')
        self.empty_fs = os.path.join(self.tempdir, 'empty.fs')
        self.server = ServerController()

    def tearDown(self):
        self.server.stop()
        shutil.rmtree(self.tempdir)
        setup.placelessTearDown()

    def test_can_create_empty_data_fs(self):
        self.server.run(self.empty_fs, '--rw')
        browser = Browser(self.server.url)
        self.assertTrue('zodbbrowser' in browser.contents)
        self.assertTrue('persistent.mapping.PersistentMapping' in browser.contents)

    def test_cannot_start_in_read_only_mode(self):
        self.assertRaises(ReadOnlyError, self.server.run, self.empty_fs)
        # Due to a bug in ZODB, the new database *is* created, it just
        # has no objects in it
        # self.assertTrue(not os.path.exists(self.empty_fs))


def printXPath(html, xpath, pretty_print=True):
    """Print a selected HTML excerpt using XPath syntax.

    Example usage:

        printXPath(browser.contents, '//div[@class="something"]')

    For more convenient integration with zope.testbrowser, you can pass
    in a browser object directly:

        printXPath(browser, '//div[@class="something"]')

    """
    printResults(html, 'xpath', xpath, pretty_print=pretty_print)


def printCSSPath(html, csspath, pretty_print=True):
    """Print a selected HTML excerpt using CSS selector syntax.

    Example usage:

        printCSS(browser.contents, 'div.something')

    For more convenient integration with zope.testbrowser, you can pass
    in a browser object directly:

        printCSS(browser, 'div.something')

    """
    printResults(html, 'cssselect', csspath, pretty_print=pretty_print)


def printResults(html, method, arg, pretty_print=True):
    if IBrowser.providedBy(html):
        # it would be nice to extract the charset from the content-type
        # header, but let's assume UTF-8, which is the only charset that
        # we use in our system
        html = html.contents.decode('UTF-8')
    # XXX: not the most appropriate place for this.  I cannot do it with
    # a renormalizer since at that point the server URL is not yet known
    html = html.replace(TestsWithServer.url, 'http://localhost/')
    results = getattr(fromstring(html), method)(arg)
    for element in results:
        if isinstance(element, basestring):
            value = element.strip()
            # XXX it would be better to specialcase lxml elements.  How?
        else:
            if pretty_print:
                fixupWhitespace(element)
            value = tostring(element, pretty_print=pretty_print).rstrip()
        if value:
            print value
    if not results:
        print "Not found: %s" % arg


def stripify(s):
    """Strip indentation and trailing whitespace from a string.

    This is a rather quirky internal function.
    """
    if s is None:
        s = ''
    had_space = s[:1].isspace()
    s = s.strip()
    if '\n' in s:
        s = ' '.join(l.strip() for l in s.splitlines())
    if had_space and s:
        s = ' ' + s
    return s


def fixupWhitespace(element, indent=0, step=2, split_if_longer=38):
    """Normalize whitespace on lxml elements."""
    # Input:
    #   <tag ...>[text]<children ...></tag>[tail]
    # Output:
    #   {indent}<tag ...>\n
    #   [indent+2][text]\n
    #             <children>
    #   [indent]</tag>\n
    #   [indent][text]

    children = element.getchildren()

    element.text = stripify(element.text)
    # heuristic for splitting long elements
    should_split = (len(str(element.attrib)) > split_if_longer or
                    len(escape(element.text)) > split_if_longer)
    if should_split and element.text:
        element.text = ('\n' + ' ' * (indent + step) + element.text.lstrip())
    if children:
        element.text += '\n' + ' ' * (indent + step)
    else:
        if '\n' in element.text:
            element.text += '\n' + ' ' * indent

    for idx, child in enumerate(children):
        fixupWhitespace(child, indent + step, step)
        if idx == len(children) - 1:
            child.tail += '\n' + ' ' * indent
        else:
            child.tail += '\n' + ' ' * (indent + step)

    if indent == 0:
        element.tail = None
    element.tail = stripify(element.tail)
    if element.tail and element.tail.startswith(' '):
        element.tail = '\n' +  ' ' * indent + element.tail.lstrip()


def setUp(test):
    test.globs['Browser'] = Browser
    test.globs['printXPath'] = printXPath
    test.globs['printCSSPath'] = printCSSPath
    test.globs['url'] = TestsWithServer.url


def test_suite():
    this = sys.modules[__name__]
    suite = unittest.defaultTestLoader.loadTestsFromModule(this)
    checker = RENormalizing([
        (re.compile(r'object at 0x[0-9a-fA-F]+'), 'object at 0xXXXXXXX'),
        (re.compile(r'\btid[0-9xA-Fa-f]+'), 'tidXXXXXXXXXXXXXXXXXX'),
        (re.compile(r'\btid=[0-9xA-Fa-f]+'), 'tid=XXXXXXXXXXXXXXXXXX'),
        (re.compile(r'\boid=[0-9xA-Fa-f]+'), 'oid=XX'),
        (re.compile(r'\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d[.]\d\d\d\d\d\d'),
            'YYYY-MM-DD HH:MM:SS.SSSSSS'),
        (re.compile(r'\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d'),
            'YYYY-MM-DD HH:MM:SS'),
        (re.compile(r'zope\.app\.folder\.folder\.Folder'),
            'zope.site.folder.Folder'),
    ])
    here = os.path.dirname(__file__)
    for filename in os.listdir(here):
        if not filename.endswith('.txt') or filename.startswith('.'):
            continue
        test = doctest.DocFileSuite(filename,
                                    setUp=setUp,
                                    checker=checker,
                                    optionflags=doctest.REPORT_NDIFF)
        test.layer = TestsWithServer
        suite.addTest(test)
    return suite




#!/usr/bin/env python
"""
ZODB Browser as a standalone app

This is a pile of hacks since bootstrapping a Zope 3 based app is incredibly
painful.
"""
import os
import stat
import sys
import asyncore
import socket
import optparse
import logging
import errno
import traceback

from ZEO.ClientStorage import ClientStorage
from ZODB.DB import DB
from ZODB.MappingStorage import MappingStorage
from ZODB.FileStorage.FileStorage import FileStorage
from ZODB.interfaces import IDatabase
from zope.app.server.servertype import IServerType
from zope.app.appsetup.appsetup import SystemConfigurationParticipation
from zope.component import getUtility, provideUtility
from zope.server.taskthreads import ThreadedTaskDispatcher
from zope.event import notify
from zope.exceptions import exceptionformatter
import zope.app.component.hooks

from zodbbrowser.state import install_provides_hack


class Options(object):
    listen_on = ('localhost', 8070)
    server_type = 'WSGI-HTTP'
    verbose = True
    threads = 4
    features = ('zserver',) # maybe 'devmode' too?
    site_definition = """
        <configure xmlns="http://namespaces.zope.org/zope"
                   i18n_domain="zodbbrowser">
          <include package="zope.securitypolicy" file="meta.zcml" />
          <include package="zope.app.zcmlfiles" file="meta.zcml" />

          <include package="zope.app.zcmlfiles" />
          <include package="zope.app.server" />
          <include package="zope.app.component" />
          <include package="zope.error"/>
          <include package="zope.publisher" />
          <include package="zope.traversing" />
          <include package="zope.traversing.browser" />

          <include package="zodbbrowser" />

          <browser:defaultView 
              for="persistent.Persistent"
              name="zodbbrowser"
              xmlns:browser="http://namespaces.zope.org/browser" />

          <unauthenticatedPrincipal id="zope.anybody" title="Unauthenticated User" />
          <unauthenticatedGroup id="zope.Anybody" title="Unauthenticated Users" />
          <authenticatedGroup id="zope.Authenticated" title="Authenticated Users" />
          <everybodyGroup id="zope.Everybody" title="All Users" />
          <securityPolicy component="zope.securitypolicy.zopepolicy.ZopeSecurityPolicy" />
          <role id="zope.Anonymous" title="Everybody"
                description="All users have this role implicitly" />
          <role id="zope.Manager" title="Site Manager" />
          <role id="zope.Member" title="Site Member" />
          <grant permission="zope.View"
                 role="zope.Anonymous" />
          <grantAll role="zope.Manager" />
          <!-- INSECURE -->
          <grant permission="zope.ManageContent"
                 role="zope.Anonymous" />
        </configure>
        """


def configure(options):
    from zope.error.error import globalErrorReportingUtility
    globalErrorReportingUtility.copy_to_zlog = True

    from zope.security.management import newInteraction, endInteraction
    endInteraction()
    newInteraction(SystemConfigurationParticipation())
    zope.app.component.hooks.setHooks()

    from zope.configuration import xmlconfig, config
    context = config.ConfigurationMachine()
    xmlconfig.registerCommonDirectives(context)
    for feature in options.features:
        context.provideFeature(feature)
    context = xmlconfig.string(options.site_definition, context=context)

    endInteraction()


def start_server(options, db):
    global task_dispatcher
    task_dispatcher = ThreadedTaskDispatcher()
    task_dispatcher.setThreadCount(options.threads)

    server = getUtility(IServerType, options.server_type)
    host, port = options.listen_on
    try:
        server.create(options.server_type, task_dispatcher, db,
                      ip=host, port=port, verbose=options.verbose)
    except socket.error, e:
        if e.errno == errno.EADDRINUSE:
            sys.exit("Cannot listen on %s:%s: %s" % (host or '0.0.0.0',
                                                     port, e))
        else:
            raise

    if options.verbose:
        print "Listening on http://%s:%d/" % (host or socket.gethostname(),
                                              port)

def serve_forever(interval=30.0):
    try:
        while asyncore.socket_map:
            asyncore.poll(interval)
    except KeyboardInterrupt:
        pass


def stop_serving():
    task_dispatcher.shutdown(False)
    asyncore.close_all()


def monkeypatch_error_formatting():
    """Use Zope's custom traceback formatter for clearer error messages."""
    traceback.format_exception = exceptionformatter.format_exception
    traceback.print_exception = exceptionformatter.print_exception


def main(args=None, start_serving=True):
    logging.basicConfig(format="%(message)s")

    if args is None:
        args = sys.argv[1:]

    parser = optparse.OptionParser(
        'usage: %prog [options] [FILENAME | --zeo ADDRESS]',
        prog='zodbbrowser',
        description='Open a ZODB database and start a web-based browser app.')
    parser.add_option('--zeo', metavar='ADDRESS',
                      help='connect to ZEO server instead'
                      ' (host:port or socket name)')
    parser.add_option('--storage', metavar='NAME',
                      help='connect to given ZEO storage')
    parser.add_option('--listen', metavar='ADDRESS',
                      help='specify port (or host:port) to listen on',
                      default='localhost:8070')
    parser.add_option('-q', '--quiet', action='store_false', dest='verbose',
                      default=True,
                      help='be quiet')
    parser.add_option('--rw', action='store_false', dest='readonly',
                      default=True,
                      help='open the database read-write (default: read-only)')
    opts, args = parser.parse_args(args)

    options = Options()
    options.verbose = opts.verbose

    if opts.listen:
        if ':' in opts.listen:
            host, port = opts.listen.rsplit(':', 1)
        else:
            host = 'localhost'
            port = opts.listen
        try:
            port = int(port)
        except ValueError:
            parser.error('invalid TCP port: %s' % port)
        options.listen_on = host, port

    if len(args) > 1:
        parser.error('too many arguments')

    if len(args) == 1:
        opts.db = args[0]
    else:
        opts.db = None

    if opts.db and opts.zeo:
        parser.error('you specified both ZEO and FileStorage; pick one')

    if opts.storage and not opts.zeo:
        parser.error('a ZEO storage was specified without ZEO connection')

    monkeypatch_error_formatting()

    if opts.db:
        filename = opts.db
        db = DB(FileStorage(filename, read_only=opts.readonly))
    elif opts.zeo:
        if ':' in opts.zeo:
            # remote hostname:port ZEO connection
            zeo_address = opts.zeo.split(':', 1)
            try:
                zeo_address[1] = int(zeo_address[1])
            except ValueError:
                parser.error('specified ZEO port must be an integer')
            zeo_address = tuple(zeo_address)
        else:
            # socket filename
            zeo_address = opts.zeo
            if os.path.exists(zeo_address):
                # try ZEO connection through UNIX socket
                mode = os.stat(zeo_address)
                if not stat.S_ISSOCK(mode.st_mode):
                    parser.error('specified file is not a valid UNIX socket')
            else:
                # remote ZEO connection
                zeo_address = (zeo_address, 8100)
        if opts.storage:
            zeo_storage = opts.storage
        else:
            zeo_storage = '1'
        db = DB(ClientStorage(zeo_address, storage=zeo_storage, read_only=opts.readonly))
    else:
        parser.error('please specify a database')

    internal_db = DB(MappingStorage())

    configure(options)

    provideUtility(db, IDatabase, name='<target>')

    notify(zope.app.appsetup.interfaces.DatabaseOpened(internal_db))

    start_server(options, internal_db)

    notify(zope.app.appsetup.interfaces.ProcessStarting())

    install_provides_hack()

    if start_serving:
        serve_forever()


if __name__ == '__main__':
    main()



import doctest

from zope.app.testing import setup
from zope.component import provideAdapter

from zodbbrowser.diff import compareDicts, compareDictsHTML
from zodbbrowser.diff import compareTuples, compareTuplesHTML
from zodbbrowser.testing import SimpleValueRenderer


def pprintDict(d):
    """Pretty-print a dictionary.

    pprint.pprint() doesn't cut it: when the dict is too short, it uses
    repr() which has no defined ordering.
    """
    print '{%s}' % ',\n '.join('%r: %r' % (key, value)
                               for key, value in sorted(d.items()))


def doctest_compareDicts():
    """Tests for compareDicts

        >>> old = dict(a=1, b=2, c=3)
        >>> new = dict(a=1, b=3, e=4)
        >>> pprintDict(compareDicts(new, old))
        {'b': ('changed to', 3, 2),
         'c': ('removed', None, 3),
         'e': ('added', 4, None)}

    """


def doctest_compareTuples():
    """Tests for compareTuples

        >>> old = (1, 2, 3)
        >>> new = ()
        >>> compareTuples(new, old)
        ((), (1, 2, 3), (), ())

        >>> old = (1, 2, 3)
        >>> new = (1, 2, 3, 4)
        >>> compareTuples(new, old)
        ((1, 2, 3), (), (4,), ())

        >>> old = (1, 2, 3, 5)
        >>> new = (2, 3, 4, 5)
        >>> compareTuples(new, old)
        ((), (1, 2, 3), (2, 3, 4), (5,))

        >>> old = (1, 2, 3, 5)
        >>> new = (1, 3, 4, 5)
        >>> compareTuples(new, old)
        ((1,), (2, 3), (3, 4), (5,))

        >>> old = (1, 2, 3, 5)
        >>> new = (1, 3, 4, 6)
        >>> compareTuples(new, old)
        ((1,), (2, 3, 5), (3, 4, 6), ())

        >>> old = (1, 2, 3, 5)
        >>> new = (2, 3, 4, 6)
        >>> compareTuples(new, old)
        ((), (1, 2, 3, 5), (2, 3, 4, 6), ())

    """


def doctest_compareTuplesHTML():
    """Tests for compareTuplesHTML

        >>> old = (1, 2, 3, 5)
        >>> new = (2, 3, 4, 5)
        >>> print compareTuplesHTML(new, old)
        <div class="diff">
          <div class="diffitem removed">
            removed 1
          </div>
          <div class="diffitem removed">
            removed 2
          </div>
          <div class="diffitem removed">
            removed 3
          </div>
          <div class="diffitem added">
            added 2
          </div>
          <div class="diffitem added">
            added 3
          </div>
          <div class="diffitem added">
            added 4
          </div>
          <div class="diffitem same">
            last item kept the same
          </div>
        </div>

        >>> old = (1, 2, 3, 5)
        >>> new = (1, 3, 4, 5)
        >>> print compareTuplesHTML(new, old)
        <div class="diff">
          <div class="diffitem same">
            first item kept the same
          </div>
          <div class="diffitem removed">
            removed 2
          </div>
          <div class="diffitem removed">
            removed 3
          </div>
          <div class="diffitem added">
            added 3
          </div>
          <div class="diffitem added">
            added 4
          </div>
          <div class="diffitem same">
            last item kept the same
          </div>
        </div>

        >>> old = (1, 2, 3, 5)
        >>> new = (1, 3, 4, 6)
        >>> print compareTuplesHTML(new, old)
        <div class="diff">
          <div class="diffitem same">
            first item kept the same
          </div>
          <div class="diffitem removed">
            removed 2
          </div>
          <div class="diffitem removed">
            removed 3
          </div>
          <div class="diffitem removed">
            removed 5
          </div>
          <div class="diffitem added">
            added 3
          </div>
          <div class="diffitem added">
            added 4
          </div>
          <div class="diffitem added">
            added 6
          </div>
        </div>

        >>> old = (1, 2, 3, 5)
        >>> new = (2, 3, 4, 6)
        >>> print compareTuplesHTML(new, old)
        <div class="diff">
          <div class="diffitem removed">
            removed 1
          </div>
          <div class="diffitem removed">
            removed 2
          </div>
          <div class="diffitem removed">
            removed 3
          </div>
          <div class="diffitem removed">
            removed 5
          </div>
          <div class="diffitem added">
            added 2
          </div>
          <div class="diffitem added">
            added 3
          </div>
          <div class="diffitem added">
            added 4
          </div>
          <div class="diffitem added">
            added 6
          </div>
        </div>

    """


def doctest_compareDictsHTML():
    """Tests for compareDicts

        >>> old = dict(a=1, b=2, c=3)
        >>> new = dict(a=1, b=3, e=4)
        >>> print compareDictsHTML(new, old)
        <div class="diff">
          <div class="diffitem changed">
            <strong>b</strong>: changed to 3
          </div>
          <div class="diffitem removed">
            <strong>c</strong>: removed 3
          </div>
          <div class="diffitem added">
            <strong>e</strong>: added 4
          </div>
        </div>

    """


def doctest_compareDictsHTML_recursive():
    """Tests for compareDicts

        >>> old = dict(a=1, b=dict(x=2, y=5), c=3, d=42, e=dict(x=42))
        >>> new = dict(a=1, b=dict(x=3, y=5), d=dict(x=42), e=42, f=dict(x=4))
        >>> print compareDictsHTML(new, old)
        <div class="diff">
          <div class="diffitem changed">
            <strong>b</strong>: dictionary changed:
            <div class="diff">
              <div class="diffitem changed">
                <strong>x</strong>: changed to 3
              </div>
            </div>
          </div>
          <div class="diffitem removed">
            <strong>c</strong>: removed 3
          </div>
          <div class="diffitem changed">
            <strong>d</strong>: changed to {'x': 42}
          </div>
          <div class="diffitem changed">
            <strong>e</strong>: changed to 42
          </div>
          <div class="diffitem added">
            <strong>f</strong>: added {'x': 4}
          </div>
        </div>

    """


def doctest_compareDictsHTML_tid_is_used():
    """Tests for compareDicts

        >>> old = dict(a=1, b=dict(x=2, y=5))
        >>> new = dict(a=2, b=dict(x=3, y=5))
        >>> print compareDictsHTML(new, old, tid=42)
        <div class="diff">
          <div class="diffitem changed">
            <strong>a</strong>: changed to 2 [tid=42]
          </div>
          <div class="diffitem changed">
            <strong>b</strong>: dictionary changed:
            <div class="diff">
              <div class="diffitem changed">
                <strong>x</strong>: changed to 3 [tid=42]
              </div>
            </div>
          </div>
        </div>

    """


def doctest_compareDictsHTML_html_quoting():
    """Tests for compareDicts

        >>> old = {}
        >>> new = {'<': 'less than'}
        >>> print compareDictsHTML(new, old)
        <div class="diff">
          <div class="diffitem added">
            <strong>&lt;</strong>: added 'less than'
          </div>
        </div>

    """


def doctest_compareDictsHTML_nonstring_keys():
    """Tests for compareDicts

        >>> old = {1: 2}
        >>> new = {1: 3}
        >>> print compareDictsHTML(new, old)
        <div class="diff">
          <div class="diffitem changed">
            <strong>1</strong>: changed to 3
          </div>
        </div>

    """


def doctest_compareDictsHTML_has_no_unicode_problems():
    r"""Tests for compareDicts

        >>> old = {u'\N{SNOWMAN}': 1, '\xFE': 2}
        >>> new = {u'\N{SNOWMAN}': 2, '\xFE': 3}
        >>> print compareDictsHTML(new, old)
        <div class="diff">
          <div class="diffitem changed">
            <strong>'\xfe'</strong>: changed to 3
          </div>
          <div class="diffitem changed">
            <strong>u'\u2603'</strong>: changed to 2
          </div>
        </div>

    """


def setUp(test):
    setup.placelessSetUp()
    provideAdapter(SimpleValueRenderer)


def tearDown(test):
    setup.placelessTearDown()


def test_suite():
    optionflags = doctest.REPORT_NDIFF|doctest.NORMALIZE_WHITESPACE
    return doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                optionflags=optionflags)

